from simple_mailer.mailer import Mailer


def test_mailer_can_send_message(smtpd):
    mailer = Mailer(port=smtpd.port, use_tls=False)
    mailer.connect()
    mailer.send_message(
        from_="me@example.com",
        to="you@example.com",
        subject="test",
        body="Hello!",
    )
    mailer.disconnect()
    assert smtpd.sent_mail



import json

import webtest
from simple_mailer.config import Config
from simple_mailer.http import Location
from simple_mailer.tests.fixtures.smtpd import SMTPServerFixture
from simple_mailer.web import get_application
from webtest import TestApp


def test_root_path_has_link_to_mailer():
    app = TestApp(get_application())
    response = app.get("/")
    assert response.status_code == 200
    data = json.loads(response.body.decode("utf8"))
    assert data["mailer"] == Config().MAILER_PATH


def test_post_urlencode_mail_app(smtpd: SMTPServerFixture):
    app = TestApp(get_application())
    response = app.post(
        "/mail", {"email": "me@example.com", "subscribe_me": True}
    )
    assert response.status_code == 200
    assert smtpd.sent_mail


def test_post_json_mail_app(smtpd: SMTPServerFixture):
    app = TestApp(get_application())
    response = app.post_json(
        "/mail", {"email": "me@example.com", "subscribe_me": True}
    )
    assert response.status_code == 200
    assert smtpd.sent_mail
    body = smtpd.sent_mail[0].body.decode("utf8")
    assert "timestamp_utc" in body


def test_post_empty_payload_is_denied():
    app = TestApp(get_application())
    try:
        app.post_json("/mail", {})
    except webtest.app.AppError as err:
        assert "400 Bad Request" in err.args[0]


def test_location_url():
    loc = Location("www.example.com", "/bar/foo")
    assert loc.https_url == "https://www.example.com/bar/foo"



class BaseSimpleMailerException(Exception):
    """Base exception for package Simple Mailer"""


class ConfigError(BaseSimpleMailerException):
    """A configuration error has occurred"""


class UnknownCaptchaProtocol(ConfigError):
    """The configured captcha protocol is unknown and therefore unsupported"""


class ContentTypeUnsupported(BaseSimpleMailerException):
    """The submitted data is formatted using an unsupported content type"""


class InvalidCaptchaResponse(BaseSimpleMailerException):
    """The captcha response key/value pair is invalid"""


class MissingCaptchaResponse(BaseSimpleMailerException):
    """The captcha response key/value pair is missing"""


class FailedCaptchaResponse(BaseSimpleMailerException):
    """The captcha response verification with the server has failed"""


class SubmittedDataInvalid(BaseSimpleMailerException):
    """The submitted data is invalid"""


class MailServerError(BaseSimpleMailerException):
    """Got an error from the remote SMTP server"""



import datetime
import json
from urllib.parse import parse_qs

import bottle
from jinja2 import Template, UndefinedError, TemplateError
from simple_mailer import exceptions
from simple_mailer.captcha import CaptchaClient
from simple_mailer.config import Config
from simple_mailer.mailer import Mailer


class Dispatcher:
    """A controller that processes incoming HTTP requests and sends mail"""

    data: dict
    metadata: dict

    def __init__(self, data=None, metadata=None):
        self._config = Config()
        self.data = {} if data is None else data
        self.metadata = {} if metadata is None else metadata

    def process_data(self) -> "Dispatcher":
        """Process the data, applying filters and manipulations appropriately
        """
        data = self.data

        fields_to_include = self._config.FIELDS_INCLUDED
        if fields_to_include:
            data = {k: v for k, v in data.items() if k in fields_to_include}
        fields_to_exclude = self._config.FIELDS_EXCLUDED
        if fields_to_exclude:
            data = {
                k: v for k, v in data.items() if k not in fields_to_exclude
            }

        if not data:
            raise exceptions.SubmittedDataInvalid("Need at least one field")
        self.data = data
        return self

    def parse_request(self, request: bottle.Request) -> "Dispatcher":
        """Extract and store the payload of a given HTTP request"""
        env = request.environ
        content_type = env["CONTENT_TYPE"]
        client_ip = env.get("HTTP_X_FORWARDED_FOR", env.get("REMOTE_ADDR", ""))
        self.metadata = {
            "mailer_url": request.url,
            "origin": request.remote_addr or "",
            "client_ip": client_ip,
        }
        body = request.body.read().decode("utf8")
        if content_type == "application/x-www-form-urlencoded":
            data = parse_qs(body)
        elif content_type == "application/json":
            try:
                data = json.loads(body)
            except json.JSONDecodeError:
                raise exceptions.SubmittedDataInvalid("Invalid JSON")
        else:
            raise exceptions.ContentTypeUnsupported(
                f"Cannot process content type: {content_type}"
            )
        self.data = data
        self.process_data()
        return self

    @property
    def template_path(self) -> str:
        """The path to the template file"""
        return self._config.MAIL_TEMPLATE_PATH

    def _get_templated_body(self) -> str:
        """Assemble and return the body of the message using the template"""
        tmpl_path = self.template_path
        if tmpl_path:
            try:
                with open(tmpl_path) as fd:
                    tmpl = Template(fd.read())
                    try:
                        return tmpl.render(
                            data=self.data, metadata=self.metadata
                        )
                    except UndefinedError as exc:
                        raise TemplateError(
                            f"The template did not define the required fields:"
                            f" {exc.message}"
                        )
            except IOError:
                raise exceptions.ConfigError(
                    f"Cannot open template file. "
                    f"Check if it exists and is readable."
                )
        else:
            return json.dumps(self.data)

    def check_captcha(self) -> None:
        """Verify that the captcha challenge was responded correctly

        If the captchas are not configured (using environment variables) this
        is a no-op."""
        captcha = CaptchaClient.from_environment()
        captcha.validate_data(self.data)

    def dispatch(self) -> None:
        """Dispatch a given HTTP request

        Returns true or false depending on the outcome.
        """
        try:
            config = Config()
        except ValueError:
            bottle.response.status_code = 500
            raise exceptions.ConfigError(
                "Mailer server application configuration error"
            )

        self.check_captcha()

        self.metadata.update(
            timestamp_utc=datetime.datetime.utcnow().isoformat()
        )

        server = Mailer(
            host=config.SMTP_HOST,
            port=config.SMTP_PORT,
            use_tls=config.USE_TLS,
        )
        server.connect()
        server.send_message(
            from_=config.MAIL_FROM,
            to=config.MAIL_TO,
            subject=config.MAIL_SUBJECT,
            body=self._get_templated_body(),
        )
        server.disconnect()



import os
from pathlib import Path

# TODO: lazy load using a metaclass?
from typing import List, Optional
from urllib.parse import urlparse

from simple_mailer.http import Location


class Config:
    def __init__(self):
        self.SMTP_HOST: str = os.environ.get("SMTP_HOST", "localhost")
        self._SMTP_PORT: str = os.environ.get("SMTP_PORT", "465")
        self.SMTP_PASSWD: str = os.environ.get("SMTP_PASSWD", "")
        self.SMTP_USERID: str = os.environ.get("SMTP_USERID", "")
        self.MAIL_TO: str = os.environ.get("TO_ADDRESS", "")
        self.MAIL_FROM: str = os.environ.get("FROM_ADDRESS", "")
        self.MAIL_SUBJECT: str = os.environ.get("MAIL_SUBJECT", "")
        self.MAILER_PATH = os.environ.get("MAILER_PATH", "/mail")
        self._USE_TLS: str = os.environ.get("USE_TLS", "true")
        self.MAIL_TEMPLATE_PATH: str = os.environ.get(
            "MAIL_TEMPLATE_PATH",
            (Path(__file__).parent / "templates" / "default.txt").resolve(),
        )
        self.CAPTCHA_TYPE: str = os.environ.get("CAPTCHA_TYPE", "")
        self.CAPTCHA_SECRET: str = os.environ.get("CAPTCHA_SECRET", "")
        self.CAPTCHA_VERIFY_URL: str = os.environ.get("CAPTCHA_VERIFY_URL", "")
        self._FIELDS_INCLUDED = os.environ.get("FIELDS_INCLUDED", "")
        self._FIELDS_EXCLUDED = os.environ.get("FIELDS_EXCLUDED", "")

    @property
    def SMTP_PORT(self) -> int:
        try:
            return int(self._SMTP_PORT)
        except ValueError:
            raise ValueError("Invalid port number.")

    @property
    def USE_TLS(self) -> bool:
        return True if self._USE_TLS.lower() == "true" else False

    @property
    def CAPTCHA_VERIFY_LOCATION(self) -> Optional[Location]:
        """The location of the captcha verification site"""
        overridden_url = self.CAPTCHA_VERIFY_URL
        if overridden_url:
            parsed = urlparse(overridden_url)
            return Location(parsed.hostname, parsed.path)
        return None

    @property
    def FIELDS_EXCLUDED(self) -> List[str]:
        """A list of fields that should be ignored

        Note: this overrides FIELDS_INCLUDED."""
        fields = (name.strip() for name in self._FIELDS_EXCLUDED.split(","))
        return [fld for fld in fields if fld]

    @property
    def FIELDS_INCLUDED(self) -> List[str]:
        """A list of fields that should be included"""
        fields = (name.strip() for name in self._FIELDS_INCLUDED.split(","))
        return [fld for fld in fields if fld]



import random
import string

from simple_mailer import captcha
from simple_mailer.dispatcher import Dispatcher
from simple_mailer.tests.helpers import with_environ_var
from simple_mailer.web import get_application
from webtest import TestApp


@with_environ_var("CAPTCHA_TYPE", "recaptchav3")
def test_captcha_interface(captcha_server, mocked_https_client):
    """Test the behaviour of the generic captcha object"""
    client = captcha.CaptchaClient.from_environment()
    assert client.validate_data({client.key: "abc"}) is None


def test_recaptcha_v3(smtpd, captcha_server, mocked_https_client):
    app = TestApp(get_application())
    recaptcha_response = "".join(
        random.choice(string.ascii_letters) for _ in range(40)
    )
    response = app.post(
        "/mail",
        {
            "email": "me@example.com",
            "subscribe_me": True,
            captcha.Recaptchav3Client.key: recaptcha_response,
        },
    )
    assert response.status_code == 200
    assert smtpd.sent_mail


def test_captcha_field_implictly_included():
    data = {
        "delete_me": "now",
        captcha.Recaptchav3Client.key: "abcdefgh1234",
    }
    assert captcha.Recaptchav3Client.key in data
    dispatcher = Dispatcher(data=data).process_data()
    assert captcha.Recaptchav3Client.key in dispatcher.data



import json

from bottle import request, response, default_app, AppStack, get, post, run
from simple_mailer import exceptions
from simple_mailer.config import Config
from simple_mailer.dispatcher import Dispatcher


@post(Config().MAILER_PATH)
def mail() -> str:
    """A resource that can send mail"""
    try:
        Dispatcher().parse_request(request).dispatch()
    except exceptions.ContentTypeUnsupported as exc:
        response.status = 400
        return str(exc)
    except exceptions.SubmittedDataInvalid as exc:
        response.status = 400
        return str(exc)
    except exceptions.BaseSimpleMailerException as exc:
        response.status = 503
        return str(exc)
    else:
        response.status = 200
        return "OK"


@get("/")
def root() -> str:
    """The root resource"""
    return json.dumps({"mailer": Config().MAILER_PATH})


def get_application() -> AppStack:
    """Get the default Bottle application"""
    return default_app()


def run_application():
    run(host="localhost", port=8080)
